/**************************************************************************//**
 * @file     spi_flash.c
 * @version  V1.00
 * $Revision: 1 $
 * $Date: 15/05/07 4:46p $
 * @brief    Read/write Flash via SPI interface
 *
 * @note
 * Copyright (C) 2015-2018 Nuvoton Technology Corp. All rights reserved.
 *
 ******************************************************************************/
#include <stdio.h>
#include <string.h>

#include "N9H30.h"
#include "sys.h"
#include "gpio.h"
#include "spi.h"

#include "GUI.h"

#define TEST_NUMBER 100   /* page numbers */
#define TEST_LENGTH 256 /* length */

static uint8_t s_SrcArray[TEST_LENGTH];
static uint8_t s_DestArray[TEST_LENGTH];

static char s_Buf[TEST_LENGTH];

void D2D3_SwitchToNormalMode(void)
{
		outpw(REG_SYS_GPB_MFPH, (inpw(REG_SYS_GPB_MFPH) & ~0xff00));
	
    /* Configure PB10 to output mode */
    GPIO_OpenBit(GPIOB, BIT10, DIR_OUTPUT, NO_PULL_UP);

    /* Configure PB11 to output mode */
    GPIO_OpenBit(GPIOB, BIT11, DIR_OUTPUT, NO_PULL_UP);

    /* Set PB.10 and PB.11 to High */
    GPIO_SetBit(GPIOB, BIT10);
		GPIO_SetBit(GPIOB, BIT11);
}

void D2D3_SwitchToQuadMode(void)
{
    outpw(REG_SYS_GPB_MFPH, (inpw(REG_SYS_GPB_MFPH) & ~0xff00) | 0xBB00);
}

void SpiFlash_ChipErase_2(void)
{
    // /CS: active
    spiIoctl(0, SPI_IOC_ENABLE_SS, SPI_SS_SS0, 0);

    // send Command: 0x06, Write enable
    spiWrite(0, 0, 0x06);
    spiIoctl(0, SPI_IOC_TRIGGER, 0, 0);
    while(spiGetBusyStatus(0));

    // /CS: de-active
    spiIoctl(0, SPI_IOC_DISABLE_SS, SPI_SS_SS0, 0);

    //////////////////////////////////////////

    // /CS: active
    spiIoctl(0, SPI_IOC_ENABLE_SS, SPI_SS_SS0, 0);

    // send Command: 0xC7, Chip Erase
    spiWrite(0, 0, 0xC7);
    spiIoctl(0, SPI_IOC_TRIGGER, 0, 0);
    while(spiGetBusyStatus(0));

    // /CS: de-active
    spiIoctl(0, SPI_IOC_DISABLE_SS, SPI_SS_SS0, 0);
}

uint8_t SpiFlash_ReadStatusReg_2(void)
{
    uint8_t u8Status;

    // /CS: active
    spiIoctl(0, SPI_IOC_ENABLE_SS, SPI_SS_SS0, 0);

    // send Command: 0x05, Read status register
    spiWrite(0, 0, 0x05);
    spiIoctl(0, SPI_IOC_TRIGGER, 0, 0);
    while(spiGetBusyStatus(0));

    // read status
    spiWrite(0, 0, 0x00);
    spiIoctl(0, SPI_IOC_TRIGGER, 0, 0);
    while(spiGetBusyStatus(0));
    u8Status = spiRead(0, 0);

    // /CS: de-active
    spiIoctl(0, SPI_IOC_DISABLE_SS, SPI_SS_SS0, 0);

    return u8Status;
}

void SpiFlash_WaitReady_2(void)
{
    uint8_t ReturnValue;

    do {
        ReturnValue = SpiFlash_ReadStatusReg_2();
        ReturnValue = ReturnValue & 1;
    } while(ReturnValue!=0); // check the BUSY bit
}

void SpiFlash_NormalPageProgram_2(uint32_t StartAddress, uint8_t *u8DataBuffer)
{
    uint32_t i = 0;

    // /CS: active
    spiIoctl(0, SPI_IOC_ENABLE_SS, SPI_SS_SS0, 0);

    // send Command: 0x06, Write enable
    spiWrite(0, 0, 0x06);
    spiIoctl(0, SPI_IOC_TRIGGER, 0, 0);
    while(spiGetBusyStatus(0));

    // /CS: de-active
    spiIoctl(0, SPI_IOC_DISABLE_SS, SPI_SS_SS0, 0);


    // /CS: active
    spiIoctl(0, SPI_IOC_ENABLE_SS, SPI_SS_SS0, 0);

    // send Command: 0x02, Page program
    spiWrite(0, 0, 0x02);
    spiIoctl(0, SPI_IOC_TRIGGER, 0, 0);
    while(spiGetBusyStatus(0));

    // send 24-bit start address
    spiWrite(0, 0, (StartAddress>>16) & 0xFF);
    spiIoctl(0, SPI_IOC_TRIGGER, 0, 0);
    while(spiGetBusyStatus(0));

    spiWrite(0, 0, (StartAddress>>8) & 0xFF);
    spiIoctl(0, SPI_IOC_TRIGGER, 0, 0);
    while(spiGetBusyStatus(0));

    spiWrite(0, 0, StartAddress & 0xFF);
    spiIoctl(0, SPI_IOC_TRIGGER, 0, 0);
    while(spiGetBusyStatus(0));

    // write data
    for(i=0; i<255; i++) {
        spiWrite(0, 0, u8DataBuffer[i]);
        spiIoctl(0, SPI_IOC_TRIGGER, 0, 0);
        while(spiGetBusyStatus(0));
    }

    // /CS: de-active
    spiIoctl(0, SPI_IOC_DISABLE_SS, SPI_SS_SS0, 0);
}

void SpiFlash_NormalRead_2(uint32_t StartAddress, uint8_t *u8DataBuffer)
{
    uint32_t i;

    // /CS: active
    spiIoctl(0, SPI_IOC_ENABLE_SS, SPI_SS_SS0, 0);

    // send Command: 0x03, Read data
    spiWrite(0, 0, 0x03);
    spiIoctl(0, SPI_IOC_TRIGGER, 0, 0);
    while(spiGetBusyStatus(0));

    // send 24-bit start address
    spiWrite(0, 0, (StartAddress>>16) & 0xFF);
    spiIoctl(0, SPI_IOC_TRIGGER, 0, 0);
    while(spiGetBusyStatus(0));

    spiWrite(0, 0, (StartAddress>>8) & 0xFF);
    spiIoctl(0, SPI_IOC_TRIGGER, 0, 0);
    while(spiGetBusyStatus(0));

    spiWrite(0, 0, StartAddress & 0xFF);
    spiIoctl(0, SPI_IOC_TRIGGER, 0, 0);
    while(spiGetBusyStatus(0));

    // read data
    for(i=0; i<256; i++) {
        spiWrite(0, 0, 0x00);
        spiIoctl(0, SPI_IOC_TRIGGER, 0, 0);
        while(spiGetBusyStatus(0));
        u8DataBuffer[i] = spiRead(0, 0);
    }
}

void SpiFlash_WriteStatusReg_2(uint8_t u8Value1, uint8_t u8Value2)
{
		// /CS: active
    spiIoctl(0, SPI_IOC_ENABLE_SS, SPI_SS_SS0, 0);

    // send Command: 0x06, Write enable
    spiWrite(0, 0, 0x06);
    spiIoctl(0, SPI_IOC_TRIGGER, 0, 0);
    while(spiGetBusyStatus(0));

		// /CS: de-active
    spiIoctl(0, SPI_IOC_DISABLE_SS, SPI_SS_SS0, 0);

		// /CS: active
    spiIoctl(0, SPI_IOC_ENABLE_SS, SPI_SS_SS0, 0);
	
		// send Command: 0x01, Write status register
    spiWrite(0, 0, 0x01);
    spiIoctl(0, SPI_IOC_TRIGGER, 0, 0);
    while(spiGetBusyStatus(0));
	
		// write status
    spiWrite(0, 0, u8Value1);
    spiIoctl(0, SPI_IOC_TRIGGER, 0, 0);
    while(spiGetBusyStatus(0));

    spiWrite(0, 0, u8Value2);
    spiIoctl(0, SPI_IOC_TRIGGER, 0, 0);
    while(spiGetBusyStatus(0));
		
		// /CS: de-active
    spiIoctl(0, SPI_IOC_DISABLE_SS, SPI_SS_SS0, 0);			
}

uint8_t SpiFlash_ReadStatusReg2_2(void)
{
    uint8_t u8Status;

		// /CS: active
    spiIoctl(0, SPI_IOC_ENABLE_SS, SPI_SS_SS0, 0);

    // send Command: 0x35, Read status register
    spiWrite(0, 0, 0x35);
    spiIoctl(0, SPI_IOC_TRIGGER, 0, 0);
    while(spiGetBusyStatus(0));

    // read status
    spiWrite(0, 0, 0x00);
    spiIoctl(0, SPI_IOC_TRIGGER, 0, 0);
    while(spiGetBusyStatus(0));
    u8Status = spiRead(0, 0);

    // /CS: de-active
    spiIoctl(0, SPI_IOC_DISABLE_SS, SPI_SS_SS0, 0);

    return u8Status;
}

void spiFlash_EnableQEBit_2(void)
{
    uint8_t u8Status1 = SpiFlash_ReadStatusReg_2();
    uint8_t u8Status2 = SpiFlash_ReadStatusReg2_2();

    u8Status2 |= 0x2;
    SpiFlash_WriteStatusReg_2(u8Status1, u8Status2);
    SpiFlash_WaitReady_2();
}

void spiFlash_DisableQEBit_2(void)
{
    uint8_t u8Status1 = SpiFlash_ReadStatusReg_2();
    uint8_t u8Status2 = SpiFlash_ReadStatusReg2_2();

    u8Status2 &= ~0x2;
    SpiFlash_WriteStatusReg_2(u8Status1, u8Status2);
    SpiFlash_WaitReady_2();
}

void SpiFlash_QuadFastRead_2(uint32_t StartAddress, uint8_t *u8DataBuffer)
{
    uint32_t i;

    // enable quad mode
    spiFlash_EnableQEBit_2();

		// /CS: active
    spiIoctl(0, SPI_IOC_ENABLE_SS, SPI_SS_SS0, 0);

    // send Command: 0xEB, Fast Read quad data
    spiWrite(0, 0, 0xEB);
    spiIoctl(0, SPI_IOC_TRIGGER, 0, 0);
    while(spiGetBusyStatus(0));
	
    // enable QSPI quad IO mode and set direction to input
    D2D3_SwitchToQuadMode();
	
		spiIoctl(0, SPI_IOC_SET_DUAL_QUAD_MODE, SPI_QUAD_MODE, 0);
		spiIoctl(0, SPI_IOC_SET_DUAL_QUAD_DIR, SPI_DUAL_QUAD_OUTPUT, 0);
		
		// send 24-bit start address
    spiWrite(0, 0, (StartAddress>>16) & 0xFF);
    spiIoctl(0, SPI_IOC_TRIGGER, 0, 0);
    while(spiGetBusyStatus(0));

    spiWrite(0, 0, (StartAddress>>8) & 0xFF);
    spiIoctl(0, SPI_IOC_TRIGGER, 0, 0);
    while(spiGetBusyStatus(0));

    spiWrite(0, 0, StartAddress & 0xFF);
    spiIoctl(0, SPI_IOC_TRIGGER, 0, 0);
    while(spiGetBusyStatus(0));

		// send 24-bit dummy byte
    spiWrite(0, 0, 0);
    spiIoctl(0, SPI_IOC_TRIGGER, 0, 0);
    while(spiGetBusyStatus(0));
		
		spiWrite(0, 0, 0);
    spiIoctl(0, SPI_IOC_TRIGGER, 0, 0);
    while(spiGetBusyStatus(0));
		
		spiWrite(0, 0, 0);
    spiIoctl(0, SPI_IOC_TRIGGER, 0, 0);
    while(spiGetBusyStatus(0));
		
		// Set Quad mode input
		spiIoctl(0, SPI_IOC_SET_DUAL_QUAD_DIR, SPI_DUAL_QUAD_INPUT, 0);
		
    // read data
    for(i=0; i<256; i++) {
        spiWrite(0, 0, 0x00);
        spiIoctl(0, SPI_IOC_TRIGGER, 0, 0);
        while(spiGetBusyStatus(0));
        u8DataBuffer[i] = spiRead(0, 0);
    }
		
		// /CS: de-active
    spiIoctl(0, SPI_IOC_DISABLE_SS, SPI_SS_SS0, 0);
		
    // Disable Quad mode
		spiIoctl(0, SPI_IOC_SET_DUAL_QUAD_MODE, SPI_DISABLE_DUAL_QUAD, 0);
		
    D2D3_SwitchToNormalMode();

    // disable quad mode
    spiFlash_DisableQEBit_2();
}


uint16_t SpiFlash_ReadMidDid_2(void)
{
    uint8_t u8RxData[2];

    // /CS: active
    spiIoctl(0, SPI_IOC_ENABLE_SS, SPI_SS_SS0, 0);

    // send Command: 0x90, Read Manufacturer/Device ID
    spiWrite(0, 0, 0x90);
    spiIoctl(0, SPI_IOC_TRIGGER, 0, 0);
    while(spiGetBusyStatus(0));

    // send 24-bit '0', dummy
    spiWrite(0, 0, 0x00);
    spiIoctl(0, SPI_IOC_TRIGGER, 0, 0);
    while(spiGetBusyStatus(0));

    spiWrite(0, 0, 0x00);
    spiIoctl(0, SPI_IOC_TRIGGER, 0, 0);
    while(spiGetBusyStatus(0));

    spiWrite(0, 0, 0x00);
    spiIoctl(0, SPI_IOC_TRIGGER, 0, 0);
    while(spiGetBusyStatus(0));

    // receive 16-bit
    spiWrite(0, 0, 0x00);
    spiIoctl(0, SPI_IOC_TRIGGER, 0, 0);
    while(spiGetBusyStatus(0));
    u8RxData[0] = spiRead(0, 0);

    spiWrite(0, 0, 0x00);
    spiIoctl(0, SPI_IOC_TRIGGER, 0, 0);
    while(spiGetBusyStatus(0));
    u8RxData[1] = spiRead(0, 0);

    // /CS: de-active
    spiIoctl(0, SPI_IOC_DISABLE_SS, SPI_SS_SS0, 0);

    return ( (u8RxData[0]<<8) | u8RxData[1] );
}

int32_t main_2(void)
{
    uint16_t u16ID;
    uint32_t u32ByteCount, u32FlashAddress, u32PageNumber;
    uint32_t nError = 0, u32Speed;

    I32       t0;
    I32       t;


//    outpw(REG_CLK_HCLKEN, 0x0527);
//    outpw(REG_CLK_PCLKEN0, 0);
//    outpw(REG_CLK_PCLKEN1, 0);

//    sysDisableCache();
//    sysFlushCache(I_D_CACHE);
//    sysEnableCache(CACHE_WRITE_BACK);
//    sysInitializeUART();

    /* Configure multi function pins to SPI0 */
    outpw(REG_SYS_GPB_MFPL, (inpw(REG_SYS_GPB_MFPL) & ~0xff000000) | 0xBB000000);
    outpw(REG_SYS_GPB_MFPH, (inpw(REG_SYS_GPB_MFPH) & ~0xff) | 0xBB);

    spiInit(0);
    spiOpen(0);

    // set spi interface speed to 2MHz
    u32Speed = 2000000;
    spiIoctl(0, SPI_IOC_SET_SPEED, u32Speed, 0);
    sprintf(s_Buf,"set spi interface speed to %d Hz\n", u32Speed);
    GUI_DispString(s_Buf);
    // set transfer length to 8-bit
    spiIoctl(0, SPI_IOC_SET_TX_BITLEN, 8, 0);
    GUI_DispString("set transfer length to 8-bit\n");
    // set transfer mode to mode-0
    spiIoctl(0, SPI_IOC_SET_MODE, 0, 0);
    GUI_DispString("set transfer mode to mode-0\n");

		D2D3_SwitchToNormalMode();
		
    // check flash id
    if((u16ID = SpiFlash_ReadMidDid_2()) == 0xEF17) {
        sysprintf("Flash found: W25Q128BV ...\n");
        GUI_DispString("Flash found: W25Q128BV ...\n");
    } else {
        sysprintf("Flash ID, 0x%x\n", u16ID);
        sprintf(s_Buf,"Flash ID, 0x%x\n", u16ID);
        GUI_DispString(s_Buf);
    }

    sysprintf("Erase chip ...");
    GUI_DispString("Erase whole chip ...");

    /* Erase SPI flash */
    SpiFlash_ChipErase_2();

    /* Wait ready */
    SpiFlash_WaitReady_2();

    sysprintf("[OK]\n");
    GUI_DispString("[OK]\n");

    /* init source data buffer */
    for(u32ByteCount=0; u32ByteCount<TEST_LENGTH; u32ByteCount++) {
        s_SrcArray[u32ByteCount] = u32ByteCount;
    }

    sysprintf("Start to normal write data to Flash ...");
    GUI_DispString("Start to normal write data to Flash ...");
    /* Program SPI flash */
    u32FlashAddress = 0;
    for(u32PageNumber=0; u32PageNumber<TEST_NUMBER; u32PageNumber++) {
        /* page program */
        SpiFlash_NormalPageProgram_2(u32FlashAddress, s_SrcArray);
        SpiFlash_WaitReady_2();
        u32FlashAddress += 0x100;
    }

    sysprintf("[OK]\n");
    GUI_DispString("[OK]\n");

    /* clear destination data buffer */
    for(u32ByteCount=0; u32ByteCount<TEST_LENGTH; u32ByteCount++) {
        s_DestArray[u32ByteCount] = 0;
    }

    sysprintf("Quad mode Read & Compare ...");
    GUI_DispString("Quad mode Read & Compare ...");

    t0 = GUI_GetTime();

    /* Read SPI flash */
    u32FlashAddress = 0;
    for(u32PageNumber=0; u32PageNumber<TEST_NUMBER; u32PageNumber++) {
        /* page read */
        SpiFlash_QuadFastRead_2(u32FlashAddress, s_DestArray);
        u32FlashAddress += 0x100;

        for(u32ByteCount=0; u32ByteCount<TEST_LENGTH; u32ByteCount++) {
            if(s_DestArray[u32ByteCount] != s_SrcArray[u32ByteCount])
                nError ++;
        }
    }

    t = GUI_GetTime();
    t -= t0;

    if(nError == 0)
    {
        sysprintf("[OK]\n");
        GUI_DispString("[OK] ");
//        sprintf(s_Buf,"Takes about %d ms\n", t);
//        GUI_DispString(s_Buf);
    }
    else
    {
        sysprintf("[FAIL]\n");
        GUI_DispString("[FAIL]\n");
    }

    return 1;
}
